@@ -51,10 +51,6 @@
 #include "util_time.h"
 #include "ap_mpm.h"
 
-#ifndef APR_LARGEFILE
-#define APR_LARGEFILE 0
-#endif
-
 typedef struct {
     char    *t_name;
     int      t_val;
@@ -143,8 +139,6 @@ static const TRANS priorities[] = {
     {NULL,      -1},
 };
 
-static apr_pool_t *stderr_pool = NULL;
-
 static apr_file_t *stderr_log = NULL;
 
 /* track pipe handles to close in child process */
@@ -157,7 +151,7 @@ static read_handle_t *read_handles;
 
 /* clear_handle_list() is called when plog is cleared; at that
  * point we need to forget about our old list of pipe read
- * handles.  We let the plog cleanups close the actual pipes.
+ * handles
  */
 static apr_status_t clear_handle_list(void *v)
 {
@@ -165,26 +159,15 @@ static apr_status_t clear_handle_list(void *v)
     return APR_SUCCESS;
 }
 
-/* remember to close this handle in the child process
- *
- * On Win32 this makes zero sense, because we don't
- * take the parent process's child procs.
- * If the win32 parent instead passed each and every
- * logger write handle from itself down to the child,
- * and the parent manages all aspects of keeping the 
- * reliable pipe log children alive, this would still
- * make no sense :)  Cripple it on Win32.
- */
+/* remember to close this handle in the child process */
 static void close_handle_in_child(apr_pool_t *p, apr_file_t *f)
 {
-#ifndef WIN32
     read_handle_t *new_handle;
 
     new_handle = apr_pcalloc(p, sizeof(read_handle_t));
     new_handle->next = read_handles;
     new_handle->handle = f;
     read_handles = new_handle;
-#endif
 }
 
 void ap_logs_child_init(apr_pool_t *p, server_rec *s)
@@ -202,7 +185,7 @@ AP_DECLARE(void) ap_open_stderr_log(apr_pool_t *p)
     apr_file_open_stderr(&stderr_log, p);
 }
 
-AP_DECLARE(apr_status_t) ap_replace_stderr_log(apr_pool_t *p, 
+AP_DECLARE(apr_status_t) ap_replace_stderr_log(apr_pool_t *p,
                                                const char *fname)
 {
     apr_file_t *stderr_file;
@@ -222,36 +205,15 @@ AP_DECLARE(apr_status_t) ap_replace_stderr_log(apr_pool_t *p,
                      ap_server_argv0, fname);
         return rc;
     }
-    if (!stderr_pool) {
-        /* This is safe provided we revert it when we are finished.
-         * We don't manager the callers pool!
-         */
-        stderr_pool = p;
-    }
-    if ((rc = apr_file_open_stderr(&stderr_log, stderr_pool)) 
-            == APR_SUCCESS) {
+    if ((rc = apr_file_open_stderr(&stderr_log, p)) == APR_SUCCESS) {
         apr_file_flush(stderr_log);
-        if ((rc = apr_file_dup2(stderr_log, stderr_file, stderr_pool)) 
-                == APR_SUCCESS) {
+        if ((rc = apr_file_dup2(stderr_log, stderr_file, p)) == APR_SUCCESS) {
             apr_file_close(stderr_file);
-            /*
-             * You might ponder why stderr_pool should survive?
-             * The trouble is, stderr_pool may have s_main->error_log,
-             * so we aren't in a position to destory stderr_pool until
-             * the next recycle.  There's also an apparent bug which 
-             * is not; if some folk decided to call this function before 
-             * the core open error logs hook, this pool won't survive.
-             * Neither does the stderr logger, so this isn't a problem.
-             */
         }
     }
-    /* Revert, see above */
-    if (stderr_pool == p)
-        stderr_pool = NULL;
-
     if (rc != APR_SUCCESS) {
         ap_log_error(APLOG_MARK, APLOG_CRIT, rc, NULL,
-                     "unable to replace stderr with error log file");
+                     "unable to replace stderr with error_log");
     }
     return rc;
 }
@@ -278,31 +240,28 @@ static int log_child(apr_pool_t *p, const char *progname,
     apr_status_t rc;
     apr_procattr_t *procattr;
     apr_proc_t *procnew;
-    apr_file_t *outfile, *errfile;
+    apr_file_t *errfile;
 
     if (((rc = apr_procattr_create(&procattr, p)) == APR_SUCCESS)
+        && ((rc = apr_procattr_cmdtype_set(procattr,
+                                           APR_SHELLCMD_ENV)) == APR_SUCCESS)
         && ((rc = apr_procattr_io_set(procattr,
                                       APR_FULL_BLOCK,
                                       APR_NO_PIPE,
                                       APR_NO_PIPE)) == APR_SUCCESS)
         && ((rc = apr_procattr_error_check_set(procattr, 1)) == APR_SUCCESS)
-        && ((rc = apr_procattr_child_errfn_set(procattr, log_child_errfn))
-                == APR_SUCCESS)) {
+        && ((rc = apr_procattr_child_errfn_set(procattr, log_child_errfn)) == APR_SUCCESS)
+        && (!dummy_stderr 
+            || ((rc = apr_file_open_stdout(&errfile, p)) == APR_SUCCESS
+                && (rc = apr_procattr_child_err_set(procattr, 
+                                                    errfile, 
+                                                    errfile)) == APR_SUCCESS))) {
         char **args;
         const char *pname;
 
         apr_tokenize_to_argv(progname, &args, p);
         pname = apr_pstrdup(p, args[0]);
         procnew = (apr_proc_t *)apr_pcalloc(p, sizeof(*procnew));
-
-        if ((rc = apr_file_open_stdout(&outfile, p)) == APR_SUCCESS) {
-            rc = apr_procattr_child_out_set(procattr, outfile, NULL);
-            if (dummy_stderr)
-                rc = apr_procattr_child_err_set(procattr, outfile, NULL);
-            else if ((rc = apr_file_open_stderr(&errfile, p)) == APR_SUCCESS)
-                rc = apr_procattr_child_err_set(procattr, errfile, NULL);
-        }
-
         rc = apr_proc_create(procnew, pname, (const char * const *)args,
                              NULL, procattr, p);
 
@@ -313,6 +272,13 @@ static int log_child(apr_pool_t *p, const char *progname,
              * close_handle_in_child()
              */
         }
+
+        /* apr_procattr_child_err_set dups errfile twice: close the
+         * remaining "parent-side" copy (apr_proc_create closes the
+         * other). */
+        if (dummy_stderr) {
+            apr_file_close(procnew->err);
+        }
     }
 
     return rc;
@@ -385,68 +351,33 @@ static int open_error_log(server_rec *s, int is_main, apr_pool_t *p)
     return OK;
 }
 
-int ap_open_logs(apr_pool_t *pconf, apr_pool_t *p /* plog */, 
+int ap_open_logs(apr_pool_t *pconf, apr_pool_t *p /* plog */,
                  apr_pool_t *ptemp, server_rec *s_main)
 {
-    apr_pool_t *stderr_p;
+    apr_status_t rc = APR_SUCCESS;
     server_rec *virt, *q;
     int replace_stderr;
+    apr_file_t *errfile = NULL;
 
-
-    /* Register to throw away the read_handles list when we
-     * cleanup plog.  Upon fork() for the apache children,
-     * this read_handles list is closed so only the parent
-     * can relaunch a lost log child.  These read handles 
-     * are always closed on exec.
-     * We won't care what happens to our stderr log child 
-     * between log phases, so we don't mind losing stderr's 
-     * read_handle a little bit early.
-     */
     apr_pool_cleanup_register(p, NULL, clear_handle_list,
                               apr_pool_cleanup_null);
-
-    /* HERE we need a stdout log that outlives plog.
-     * We *presume* the parent of plog is a process 
-     * or global pool which spans server restarts.
-     * Create our stderr_pool as a child of the plog's
-     * parent pool.
-     */
-    apr_pool_create(&stderr_p, apr_pool_parent_get(p));
-    apr_pool_tag(stderr_p, "stderr_pool");
-    
-    if (open_error_log(s_main, 1, stderr_p) != OK) {
+    if (open_error_log(s_main, 1, p) != OK) {
         return DONE;
     }
 
     replace_stderr = 1;
     if (s_main->error_log) {
-        apr_status_t rv;
-        
-        /* Replace existing stderr with new log. */
+        /* replace stderr with this new log */
         apr_file_flush(s_main->error_log);
-        rv = apr_file_dup2(stderr_log, s_main->error_log, stderr_p);
-        if (rv != APR_SUCCESS) {
-            ap_log_error(APLOG_MARK, APLOG_CRIT, rv, s_main,
+        if ((rc = apr_file_open_stderr(&errfile, p)) == APR_SUCCESS) {
+            rc = apr_file_dup2(errfile, s_main->error_log, p);
+        }
+        if (rc != APR_SUCCESS) {
+            ap_log_error(APLOG_MARK, APLOG_CRIT, rc, s_main,
                          "unable to replace stderr with error_log");
         }
         else {
-            /* We are done with stderr_pool, close it, killing
-             * the previous generation's stderr logger
-             */
-            if (stderr_pool)
-                apr_pool_destroy(stderr_pool);
-            stderr_pool = stderr_p;
             replace_stderr = 0;
-            /*
-             * Now that we have dup'ed s_main->error_log to stderr_log
-             * close it and set s_main->error_log to stderr_log. This avoids
-             * this fd being inherited by the next piped logger who would
-             * keep open the writing end of the pipe that this one uses
-             * as stdin. This in turn would prevent the piped logger from
-             * exiting.
-             */
-             apr_file_close(s_main->error_log);
-             s_main->error_log = stderr_log;
         }
     }
     /* note that stderr may still need to be replaced with something
@@ -771,7 +702,7 @@ AP_DECLARE(void) ap_log_pid(apr_pool_t *p, const char *filename)
 
     fname = ap_server_root_relative(p, filename);
     if (!fname) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, APR_EBADPATH, 
+        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, APR_EBADPATH,
                      NULL, "Invalid PID file path %s, ignoring.", filename);
         return;
     }
@@ -787,9 +718,9 @@ AP_DECLARE(void) ap_log_pid(apr_pool_t *p, const char *filename)
          *      based on the last modification time of the pid file.
          */
         ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, p,
-                      apr_psprintf(p, "pid file %s overwritten -- Unclean "
-                                   "shutdown of previous Apache run?",
-                                   fname));
+                      "pid file %s overwritten -- Unclean "
+                      "shutdown of previous Apache run?",
+                      fname);
     }
 
     if ((rv = apr_file_open(&pid_file, fname,
@@ -824,7 +755,7 @@ AP_DECLARE(apr_status_t) ap_read_pid(apr_pool_t *p, const char *filename,
 
     fname = ap_server_root_relative(p, filename);
     if (!fname) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, APR_EBADPATH, 
+        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, APR_EBADPATH,
                      NULL, "Invalid PID file path %s, ignoring.", filename);
         return APR_EGENERAL;
     }
@@ -834,9 +765,7 @@ AP_DECLARE(apr_status_t) ap_read_pid(apr_pool_t *p, const char *filename,
         return rv;
     }
 
-    /* Ensure null-termination, so that strtol doesn't go crazy. */
     buf = apr_palloc(p, BUFFER_SIZE);
-    buf[BUFFER_SIZE - 1] = '\0';
 
     rv = apr_file_read_full(pid_file, buf, BUFFER_SIZE - 1, &bytes_read);
     if (rv != APR_SUCCESS && rv != APR_EOF) {
@@ -845,10 +774,11 @@ AP_DECLARE(apr_status_t) ap_read_pid(apr_pool_t *p, const char *filename,
 
     /* If we fill the buffer, we're probably reading a corrupt pid file.
      * To be nice, let's also ensure the first char is a digit. */
-    if (bytes_read == BUFFER_SIZE - 1 || !apr_isdigit(*buf)) {
+    if (bytes_read == 0 || bytes_read == BUFFER_SIZE - 1 || !apr_isdigit(*buf)) {
         return APR_EGENERAL;
     }
 
+    buf[bytes_read] = '\0';
     *mypid = strtol(buf, &endptr, 10);
 
     apr_file_close(pid_file);
@@ -886,6 +816,8 @@ static apr_status_t piped_log_spawn(piped_log *pl)
     apr_status_t status;
 
     if (((status = apr_procattr_create(&procattr, pl->p)) != APR_SUCCESS) ||
+        ((status = apr_procattr_cmdtype_set(procattr,
+                                            APR_SHELLCMD_ENV)) != APR_SUCCESS) ||
         ((status = apr_procattr_child_in_set(procattr,
                                              ap_piped_log_read_fd(pl),
                                              ap_piped_log_write_fd(pl)))
@@ -902,12 +834,6 @@ static apr_status_t piped_log_spawn(piped_log *pl)
     else {
         char **args;
         const char *pname;
-        apr_file_t *outfile, *errfile;
-
-        if ((status = apr_file_open_stdout(&outfile, pl->p)) == APR_SUCCESS)
-            status = apr_procattr_child_out_set(procattr, outfile, NULL);
-        if ((status = apr_file_open_stderr(&errfile, pl->p)) == APR_SUCCESS)
-            status = apr_procattr_child_err_set(procattr, errfile, NULL);
 
         apr_tokenize_to_argv(pl->program, &args, pl->p);
         pname = apr_pstrdup(pl->p, args[0]);
